/*
 *
 *  D-Bus introspection utility
 *
 *  Copyright (C) 2007-2008  Intel Corporation. All rights reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>

#include <dbus/dbus.h>
#include <glib.h>

static void introspect(DBusConnection *conn,
				const char *service, const char *path);

typedef struct {
	DBusConnection *conn;
	gchar *service;
	gchar *path;
} ContextData;

static void element_start(GMarkupParseContext *context, const gchar *elem_name,
			const gchar **attr_names, const gchar **attr_values,
					gpointer user_data, GError **error)
{
	ContextData *ctx_data = user_data;
	gchar *path;

	if (attr_names == NULL || attr_names[0] == NULL)
		return;

	if (g_str_equal(attr_names[0], "name") == FALSE)
		return;

	if (g_str_equal(elem_name, "node") == TRUE) {
		if (g_str_equal(attr_values[0], ctx_data->path) == FALSE) {
			path = g_strdup_printf("%s/%s",
				g_str_equal(ctx_data->path, "/") == TRUE ?
							"" : ctx_data->path,
							attr_values[0]);
			introspect(ctx_data->conn, ctx_data->service, path);
			g_free(path);
		}
	} else if (g_str_equal(elem_name, "interface") == TRUE) {
		printf("        %s\n", attr_values[0]);
	} else if (g_str_equal(elem_name, "method") == TRUE) {
		printf("            Method: %s\n", attr_values[0]);
	} else if (g_str_equal(elem_name, "signal") == TRUE) {
		printf("            Signal: %s\n", attr_values[0]);
	} else if (g_str_equal(elem_name, "property") == TRUE) {
		printf("            Property: %s\n", attr_values[0]);
	}
}

static void element_end(GMarkupParseContext *context, const gchar *elem_name,
					gpointer user_data, GError **error)
{
}

static GMarkupParser parser = {
	element_start, element_end, NULL, NULL, NULL
};

static void parse_introspection(DBusConnection *conn, const char *service,
				const char *path, const char *introspection)
{
	GMarkupParseContext *ctx;
	ContextData *ctx_data;

	ctx_data = g_new0(ContextData, 1);
	ctx_data->conn = conn;
	ctx_data->path = g_strdup(path);
	ctx_data->service = g_strdup(service);

	ctx = g_markup_parse_context_new(&parser, 0, ctx_data, NULL);

	if (g_markup_parse_context_parse(ctx,
			introspection, strlen(introspection), NULL) == FALSE) {
		printf("    Wrong introspection data\n");
	}

	g_markup_parse_context_free(ctx);

	g_free(ctx_data->service);
	g_free(ctx_data->path);
	g_free(ctx_data);
}

static void introspect(DBusConnection *conn,
				const char *service, const char *path)
{
	DBusMessage *msg, *reply;
	const char *str;

	msg = dbus_message_new_method_call(service, path,
				DBUS_INTERFACE_INTROSPECTABLE, "Introspect");
	if (msg == NULL) {
		fprintf(stderr, "Can't allocate new method call\n");
		return;
	}

	dbus_message_append_args(msg, DBUS_TYPE_INVALID);

	reply = dbus_connection_send_with_reply_and_block(conn, msg, -1, NULL);

	dbus_message_unref(msg);

	if (reply == NULL) {
		printf("    No introspection\n");
		return;
	}

	printf("    %s\n", path);

	if (dbus_message_get_args(reply, NULL, DBUS_TYPE_STRING, &str,
						DBUS_TYPE_INVALID) == FALSE) {
		fprintf(stderr, "Can't read introspection data\n");
		goto done;
	}

	parse_introspection(conn, service, path, str);

done:
	dbus_message_unref(reply);

	dbus_connection_flush(conn);
}

static void list_names(DBusConnection *conn)
{
	DBusMessage *msg, *reply;
	const char **names;
	int i, count;

	msg = dbus_message_new_method_call(DBUS_SERVICE_DBUS, DBUS_PATH_DBUS,
					DBUS_INTERFACE_DBUS, "ListNames");
	if (msg == NULL) {
		fprintf(stderr, "Can't allocate new method call\n");
		return;
	}

	dbus_message_append_args(msg, DBUS_TYPE_INVALID);

	reply = dbus_connection_send_with_reply_and_block(conn, msg, -1, NULL);

	dbus_message_unref(msg);

	if (reply == NULL) {
		fprintf(stderr, "Can't execute introspection method\n");
		return;
	}

	if (dbus_message_get_args(reply, NULL,
				DBUS_TYPE_ARRAY, DBUS_TYPE_STRING,
				&names, &count, DBUS_TYPE_INVALID) == FALSE) {
		fprintf(stderr, "Can't read name list\n");
		goto done;
	}

	for (i = 0; i < count; i++) {
		if (*names[i] == ':')
			continue;

		printf("%s\n", names[i]);
		introspect(conn, names[i], "/");
	}

done:
	dbus_message_unref(reply);

	dbus_connection_flush(conn);
}

static void usage(void)
{
	printf("D-Bus introspection utility %s\n\n", VERSION);

	printf("Usage:\n"
		"\tintrospect [--system | --session]\n"
		"\n");
}

static struct option options[] = {
	{ "system",  0, 0, 's' },
	{ "session", 0, 0, 'u' },
	{ "path",    1, 0, 'p' },
	{ "help",    0, 0, 'h' },
	{ 0, 0, 0, 0 }
};

int main(int argc, char *argv[])
{
	DBusConnection *conn;
	DBusBusType type = DBUS_BUS_SESSION;
	char *path = NULL;
	int opt;

	while ((opt = getopt_long(argc, argv, "+sup:h", options, NULL)) != EOF) {
		switch(opt) {
		case 's':
			type = DBUS_BUS_SYSTEM;
			break;
		case 'u':
			type = DBUS_BUS_SESSION;
			break;
		case 'p':
			path = strdup(optarg);
			break;
		case 'h':
		default:
			usage();
			exit(0);
		}
	}

	argc -= optind;
	argv += optind;
	optind = 0;

	conn = dbus_bus_get(type, NULL);
	if (conn == NULL) {
		fprintf(stderr, "Can't get on the bus\n");
		exit(1);
	}

	if (argc > 0) {
		printf("%s\n", argv[0]);
		introspect(conn, argv[0], path ? path : "/");
	} else
		list_names(conn);

	if (path)
		free(path);

	dbus_connection_unref(conn);

	return 0;
}
